/* FOR COMPILATION AND DEBUGGING PURPOSES ONLY! */
/* This ai.h file contains headers for all AI routines used by the universe.
 * This can be taken as a definition of the current AI/UN API. */

/* API change: Altered function calls for generation of new genomes, to allow
 * for two parents of a probe or player, and zero parents of any other
 * object. */

#ifndef AI_H_
#define AI_H_

#include "untypes.h" /* For UN_Object */
#include "fe.h"      /* For TicksLeft */

/* Description:
 * Interrupts for failed actions, used as parameters to AI_Interrupt. */

#define THRUST_FAIL 0
#define BRAKE_FAIL 1
#define LEFT_FAIL 2
#define RIGHT_FAIL 3
#define FIRE_FAIL 4

/*****************************
 * Genome Struct
 ****************************/

typedef struct AI_Genome_s {
   long temp;
} AI_Genome;

/*****************************
 * Function Prototypes
 ****************************/

/* Description:
 * Initialize the genome of an object, whether it has parents (as is the
 * case for player objects or probe objects) or not (as is the case for
 * all other objects).  This includes setting up any internal markers for AI
 * tracking of each object, and this will be called after the object is
 * initialized (although for probe and player objects, AI will probably want
 * to overwrite the default settings for the object's values).
*/
void AI_InitProbeGenome(UN_Object* obj, AI_Genome *parent1, AI_Genome *parent2);
void AI_InitOtherGenome(UN_Object* obj);

/* Description:
 * Gets the keys of the probe. An action is currently
 * a long with each bit reprsenting a key e.g. THRUST | LEFT | FIRE
 * (see universe.h for the #defines)
 */
UN_Actions AI_GetKeys(UN_Object *obj);

/* Description:
 * Tells the probe something e.g. THRUST_FAIL, FIRE_FAIL
 */
void AI_Interrupt(UN_Object *obj, long type);

/* Description:
 * Does the actual AI stuff for all the probes
 */
void AI_DoQueue(FE_Ticks ticksLeft);

#endif
