/* FOR COMPILATION AND DEBUGGING PURPOSES ONLY! */
/* This fe.h file contains headers for all FE routines used by the universe.
 * This can be taken as a definition of the current FE/UN API. */

#ifndef FE_H
#define FE_H

#include "untypes.h"

#define TARGET_RADIUS 100

typedef unsigned short FE_ErrorNum;
enum {
	E_NOERROR = 0,
	E_INITSYS,
	E_MEMORY
	};
typedef unsigned long FE_Ticks;

typedef unsigned short FE_ScreenNum;
enum {
   MAINMENU=0,
   PAUSEMENU,
   CONFIRMQUITDIALOG,
   LOADGAMESCREEN,
   SAVEGAMESCREEN,
   CONFIRMLOAD,
   CONFIRMLOAD1, /* CONFIRMLOADn is called when the player has requested 
   CONFIRMLOAD2,  * that game n be loaded; CONFIRMLOAD by itself is a place
   CONFIRMLOAD3,  * holder to allow easy calculation of the screen to call
   CONFIRMLOAD4,  * (ie, CONFIRMLOAD + game numer).  The reason for this is
   CONFIRMLOAD5,  * because FE may wish to display details on the specific
   CONFIRMLOAD6,  * slot being loaded. */
   CONFIRMLOAD7,
   CONFIRMLOAD8,
   CONFIRMSAVE,
   CONFIRMSAVE1, /* See CONFIRMLOADn above; CONFIRMSAVEn is a similar system
   CONFIRMSAVE2,  * with a similar purpose. */
   CONFIRMSAVE3,
   CONFIRMSAVE4,
   CONFIRMSAVE5,
   CONFIRMSAVE6,
   CONFIRMSAVE7,
   CONFIRMSAVE8,
   OPTIONSCREEN
};

/* A wrapper for malloc(), which should include free memory checking */
void *FE_Malloc(unsigned long mem);

/* A wrapper for free() */
void FE_Free(void *ptr);

/* Initialize the system */
FE_ErrorNum FE_InitSystem(int *argc_ret, char *argv[]);

/* Deinitialize the system */
void FE_DeInitSystem();

/* Compile a sprite - used by AS */
char *FE_CompileSprite(char *source);

/* Return the keys currently being pressed */
UN_Actions FE_GetKeys();

/* Determine how many ticks are left until the next screen update, for use
 * by AI */
FE_Ticks FE_GetTicks(FE_Ticks nextUpdate);

/* As the title suggests.  We may want to break this up into ToggleMusic and
 * ToggleSFX. */
void FE_ToggleSound();

/* Update the screen, only checking subsectors around the given center to see
 * if the objects in the subsector should go on the screen */
FE_Ticks FE_Update(UN_Object* **Sectors,
	UN_Coord centerX, UN_Coord centerY,
	unsigned short shiftsX, unsigned short shiftsY);

/* Return the current pointer status, for use by MA menuing functions */
void FE_PointerStatus(short* x, short* y, char* b1, char* b2);

/* Display an error */
void FE_DisplayError(FE_ErrorNum err);

/* Display a pregenerated menu screen */
void FE_DisplayScreen(FE_ScreenNum screenNum);

/* Load and save the games; slot is the game number (1 through 8) */
FE_ErrorNum FE_LoadGame(short slot);
FE_ErrorNum FE_SaveGame(short slot);

/* Determine when the next screen update will be */
FE_Ticks FE_Set();

#endif
