;*********************************************************************
; FEASM.ASM
;	This contains the ASM functions required by FEPCX.
;   Compiles under TASM 3.0 with the switches specified in the makefile.
;*********************************************************************
;     Copyright 1995 by Jeff Lait.
;     Permission granted to members of the A-Life team to hack/modify at
; their whim :>
;*********************************************************************
; Notes:
;   This current version is for 320x200.  It basically does all writes to
; an offscreen buffer which is then blitted over via BlitScreen.
;   A VESA version will probably use page flipping and do all writes directly
; on screen to reduce bandwidth...
;   The timer isn't yet hooked in.
;*********************************************************************


		.386p

;*********************************************************************
; Functions:
;*********************************************************************
public FEASM_DrawSquare_	; dest, width, height.
public FEASM_Init_			; x, ylookup, keytable.
public FEASM_Shutdown_
public FEASM_SetGraph_		; int mode
public FEASM_BlitScreen_	; int dest, int source, int skip

; 3761 FPS for 32x32 trans.
public FEASM_DrawRotated_	; int dest, int source, int *edgedata
public FEASM_MoveStars_		; int dest,int numstars,int *stardata
public FEASM_InitStars_		; int seed, int num, int *data

_DATA segment para public use32 'DATA'

;*********************************************************************
; EQUATES
;*********************************************************************
SPRITEW = 32
SPRITEH = 32
VIRTW = 12
VIRTH = 9
SPRITESZ = 1024

;*********************************************************************
; DATA
;*********************************************************************
keytable DD ?
xlookup DD ?
ylookup DD ?
OldKeyOff DD ?
OldKeySeg DW ?
Random DD 12345678h

;*********************************************************************
; Jump table used by unrolled loop:
;*********************************************************************
JumpTable520:
DD offset Jump0Table520
DD offset Jump1Table520
DD offset Jump2Table520
DD offset Jump3Table520

;*********************************************************************
; Multiplication lookup.  Hey, there must be a better way to generate
; this under TASM 3.0 - Any suggestions???
;*********************************************************************
DDMultTable:		; Probably better way to write this...
IRP @M41, <0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31>
	IRP @M42, <0, 1, 2, 3, 4, 5, 6, 7, 8>
		DD VIRTW*SPRITEW*(@M41*9 + @M42)
	ENDM
ENDM

_DATA ends

;*********************************************************************
; CODE
;*********************************************************************

_TEXT segment para public use32 'CODE'
		assume cs:_TEXT, ds:_DATA, es:_DATA

;*********************************************************************
; FUNCTIONS
;*********************************************************************

;*********************************************************************
; RandomDelta
;  Sets EAX to a random delta.
; This generator was ripped from BC++ 3.0
;*********************************************************************
RandomDelta proc
	push edx ecx ebx
	mov ecx, 15a4e35h
	mov eax, Random
	mul ecx
	shrd eax, edx, 16
	inc eax
	mov ebx, eax		; Top 16 random...
	mul ecx
	shrd eax, edx, 16
	inc eax
	mov Random, eax
	shrd ebx, edx, 16
	mov eax, ebx
	pop ebx ecx edx
	ret
RandomDelta endp

;*********************************************************************
; FEASM_SetGraph - Called with
;	v FEASM_SetGraph(int mode);
;					  EAX
; TBD: Make VESA compliant.
;*********************************************************************
FEASM_SetGraph_ proc
	int 10h
	ret
FEASM_SetGraph_ endp

;*********************************************************************
; FEASM_Init - Called with
;	v FEASM_Init(char *x, char *ylookup);
;				  EAX			EDX
;*********************************************************************
FEASM_Init_ proc
	mov xlookup, eax
	mov ylookup, edx
; Done...
	ret
FEASM_Init_ endp

;*********************************************************************
; FEASM_Shutdown - Called with
;	v FEASM_Shutdown();
;*********************************************************************
FEASM_Shutdown_ proc
	ret
FEASM_Shutdown_ endp


;*********************************************************************
; FEASM_DrawSquare - Called as
;	v FEASM_DrawSquare(char *dest, int w, int h);
;						 EAX	    EDX    EBX
;  Fills in a black rectangle centred at dest, of wxh dimmensions.
;*********************************************************************
FEASM_DrawSquare_ proc
	push ecx edi
	mov edi, eax
	add edx, 3
	shr edx, 1
	sub edi, edx
	shr edx, 1
	mov eax, ebx
 	shr eax, 1
	sub edi, [eax*4 + offset DDMultTable]
	xor eax, eax
 	StartLoop231:
		push edi		; Ugh...
		mov ecx, edx
		rep stosd
		pop edi
		add edi, VIRTW*SPRITEW
		dec ebx
		jnz StartLoop231
	pop edi ecx
	ret
FEASM_DrawSquare_ endp

;*********************************************************************
; FEASM_BlitScreen - called with
;  v FEASM_BlitScreen(char *dest, char *source, int skip);
;                        EAX         EDX         EBX
;*********************************************************************
FEASM_BlitScreen_ proc
	push esi edi ecx
	mov edi, eax
	mov esi, edx
	mov eax, 200		; Should be made generic.
	Loop225:
		mov ecx, 80
		rep movsd
		add esi, ebx
		dec eax
		jnz Loop225
	pop ecx edi esi
	ret
FEASM_BlitScreen_ endp

;*********************************************************************
; FEASM_InitStars(int seed, int numstar, int *stardata);
;                  EAX           EDX         EBX
;  Randomizes positions & speeds of all stars.
;*********************************************************************
FEASM_InitStars_ proc
	push esi edi ecx
	mov ecx, edx
	mov esi, ebx
	add esi, 8
	mov Random, eax
	StartLoop292:
		call RandomDelta
		xor edx, edx
		mov ebx, VIRTW*SPRITEW*65536
		div ebx
		mov [esi], edx			; X
		call RandomDelta
		xor edx, edx
		mov ebx, VIRTH*SPRITEW*65536
		div ebx
		mov [esi+4], edx		; Y
		call RandomDelta
		and eax, 03ffffh
		mov [esi+8], eax		; Speed
		shr eax, 14
		add eax, 16
		mov [esi+12], eax		; Colour.
		add esi, 16
		dec ecx
		jnz StartLoop292
	pop ecx edi esi
	ret
FEASM_InitStars_ endp

;*********************************************************************
; FEASM_MoveStars(int dest, int numstar, int *stardata);
;                  EAX        EDX              EBX
; Star Format:
;   Header: DD dX, dY
;   Element:DD X, Y, Speed, Colour
; NB: Not perfect, a trailing star can be masked by a leading one, but
; to avoid this would take two passes which would probably not be worth it.
; Problems in VESA: Keep stars in seperate banks: Crossovers should be throwns
; on a crossover stack for dealing with afterwards.
;*********************************************************************
FEASM_MoveStars_ proc
	push esi edi ecx ebp
	mov edi, eax
	mov esi, ebx
	mov ecx, edx
	mov ebx, [esi]			; dX
	mov ebp, [esi+4]		; dY
	add esi, 8
	jmp LoopTest265
	StartLoop266:
		mov edx, [esi+4]	; Y
		mov eax, [esi]		; X
		shr edx, 16
		shr eax, 16
		add eax, [edx*4 + offset DDMultTable]	; Offset of star... 
		mov BYTE PTR [edi+eax], 0	; Clear star... (AGI!)
		; Update star pos:
		; TBD: Optimize for jmp pred. if req'd.
		mov eax, [esi+8]	; Speed.
		imul ebx			; dx...
		shld edx, eax, 16	; Actual dx...
		add edx, [esi]		; X
		jns Skip282
			; Negative:
			add edx, VIRTW*SPRITEW*65536
			call RandomDelta
			and eax, 03ffffh
			mov [esi+8], eax		; New dY.
			shr eax, 14				; Brighter: closer. Gives 0-16.
			add eax, 16
			mov [esi+12], eax		; New colour.
		Skip282:
		cmp edx, VIRTW*SPRITEW*65536
		jb Skip287
			sub edx, VIRTW*SPRITEW*65536
			call RandomDelta
			and eax, 03ffffh
			mov [esi+8], eax		; New dY.
			shr eax, 14				; Brighter: closer. Gives 0-16.
			add eax, 16
			mov [esi+12], eax		; New colour.
		Skip287:
		mov [esi], edx			; Save X...
		; Now Y:
		mov eax, [esi+8]		; Speed.
		imul ebp				; dY
		shld edx, eax, 16
		add edx, [esi+4]			; Y.
		jns Skip302
			add edx, VIRTH*SPRITEH*65536
			call RandomDelta
			and eax, 03ffffh
			mov [esi+8], eax		; New dY.
			shr eax, 14				; Brighter: closer. Gives 0-16.
			add eax, 16
			mov [esi+12], eax		; New colour.
		Skip302:
		cmp edx, VIRTH*SPRITEH*65536
		jb Skip311
			sub edx, VIRTH*SPRITEH*65536
			call RandomDelta
			and eax, 03ffffh
			mov [esi+8], eax		; New dY.
			shr eax, 14				; Brighter: closer. Gives 0-16.
			add eax, 16
			mov [esi+12], eax		; New colour.
		Skip311:
		mov [esi+4], edx
		; Write new star at current pos:
		mov eax, [esi]
		mov edx, [esi+4]
		shr edx, 16
		shr eax, 16
		add eax, [edx*4 + offset DDMultTable]
		mov edx, [esi+12]
		mov BYTE PTR [edi+eax], dl
		; Finish:
		add esi, 16
	LoopTest265:
		dec ecx
		jnz StartLoop266
	pop ebp ecx edi esi
	ret
FEASM_MoveStars_ endp

;*********************************************************************
; FEASM_DrawRotated - called with
;	v FEASM_DrawRotated(int dest, int source, int *edgedata
;						 EAX		EDX				EBX
; Does a simplified t-mapper using precompiled edge data.
; There appears to be an offby one error either in here or in the
; edge data compiler...
;   Times are measured in 32x32 ships per 20 seconds.
;  Not Unrolled, w/ pushes: 117053.
;  Unrolled by 4, w/ pushes: 130657.
;  Unrolled by 4, removed transparcency jump: ~43000: Ouch!!!
;*********************************************************************
FEASM_DrawRotated_ proc
	push esi edi ecx ebp

	; Set up:
	; Set proper destination:
	mov edi, eax
	mov eax, [ebx+4]	; Height
	shr eax, 1
	mov eax, [eax*4 + offset DDMultTable]
	add eax, eax
	add eax, [ebx]		; Width
	shr eax, 1
	sub edi, eax
	; Set up source:
	mov esi, edx
	; Set up increments:
	mov ax, [ebx+10]
	cwde
	mov ebp, eax		; High dxx.
	mov eax, [ebx+6]	; Load high eax with low cos.
	mov ax, [ebx+13]	; ah = lowhigh byte of sin, al = high low.
	; Get past header.
	add ebx, 16
	; Set up y loop vars:
	mov ecx, [ebx-12]	; Height.
; Argh.. This loop is disgustingly unefficient...
	StartLoop290:
		push edi esi ebx ecx	; Should store or S-M or something else.
		add esi, [ebx]
		add edi, [ebx+8]
		mov ecx, [ebx+12]
		mov edx, ecx
		add ecx, 3
		and edx, 3
		shr ecx, 2
		jmp DWORD PTR [edx*4 + offset JumpTable520]

Jump0Table520:
		mov edx, [ebx+4]
		xor ebx, ebx
		jcxz End314				; Occasionaly, really is zero!
		jmp Unrolled0Loop275
Jump1Table520:
		mov edx, [ebx+4]
		xor ebx, ebx
		sub edi, 3
		jmp Unrolled1Loop275
Jump2Table520:
		mov edx, [ebx+4]
		xor ebx, ebx
		sub edi, 2
		jmp Unrolled2Loop275
Jump3Table520:
		mov edx, [ebx+4]
		xor ebx, ebx
		dec edi
		jmp Unrolled3Loop275

		StartLoop275:
Unrolled0Loop275:
			add edx, eax
			mov bh, [esi+ebx]		; No AGI - HAHAHA!!!
			adc esi, ebp				   
			or bh, bh
			jz Skip270
				mov [edi], bh
			Skip270:
			mov bh, dh
Unrolled3Loop275:
			add edx, eax
			mov bh, [esi+ebx]		; No AGI.
			adc esi, ebp
			or bh, bh
			jz Skip566
				mov [edi+1], bh
			Skip566:
			mov bh, dh
Unrolled2Loop275:
			add edx, eax
			mov bh, [esi+ebx]		; No AGI.
			adc esi, ebp
			or bh, bh
			jz Skip575
				mov [edi+2], bh
			Skip575:
			mov bh, dh
Unrolled1Loop275:
			add edx, eax
			mov bh, [esi+ebx]		; No AGI.
			adc esi, ebp
			or bh, bh
			jz Skip584
				mov [edi+3], bh
			Skip584:
			mov bh, dh
; End...
			add edi, 4
			dec ecx
			jnz StartLoop275
		End314:
		pop ecx ebx esi edi
		add ebx, 16
		add edi, VIRTW*SPRITEW
;		add edi, 320
		dec ecx
		jnz StartLoop290
	; Should now be done...
	pop ebp ecx edi esi
	ret
FEASM_DrawRotated_ endp

ends _TEXT
end
