/******************************************************************

	$Header$

	Module: FEMAIN.CPP

	Author:	Jeff Lait

	Copyright 1995 Project Von N*.

	Description: This is all the FE-PCX code which can still be written 
				in C due to it's lack of need for speed but still must 
				be FE because it is implementation specific.
				Note that C++ style comments are used.  Other than that,
				it should work on any flat memory model compiler for 80x86.

 ******************************************************************/

/******************************************************************

  Revision Record

	Rev	Date		Auth	Changes
	===	====		====	=======

	0.0	23/10/95	jml  	created header.
	0.1 29/11/95	jml		merged in stubs from Stuart McDonald

 ******************************************************************/

/******************************************************************
 * INCLUDES:
 ******************************************************************/
#include <malloc.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include "feasm.h"		// Asm funcs specific to this.
#include "ai.h"		// alife header.
#include "main.h"		// Main data.
#include "as.h"			// artsound header.
#include "unal.h"
#include "ungame.h"
#include "unobject.h"
#include "untypes.h"	
#include "fe.h"			// Standard FE header.
#include "\generic\keyhndlr.h"
#include "\generic\joystick.h"
#include "\generic\timer.h"

/******************************************************************
 * DEFINES:
 ******************************************************************/
// These defines will have to be made generic:
#define SPRITEW 32		// `Tile size', currently onl used to determine
#define SPRITEH 32		// how much of an overhang to do on edges of screen.
#define VIRTW   12		// dimmensions of virtual screen in tiles.
#define VIRTH	9
#define SPRITESZ	SPRITEW*SPRITEH	// Size of one tile in bytes.
#define DEG_TO_RAD 3.1415 / 128
#define max(a,b) ((a<b) ? (b) : (a))
#define min(a,b) ((a>b) ? (b) : (a))
#define abs(a)   ((a < 0) ? (-(a)) : (a))
#define UNFIX(a) ((a)>>6)		// Shouldn't be here :>

#define JOYBOUND 100

// POINT_IN: Determines whether a point is inside square with top left at (0,0)
#define POINT_IN(x, y, w, h) \
	((x > 0) && (x < w) && (y > 0) && (y < h))

#define NUM_STAR 500			// Make as large as desirable.
#define STATE_Game
// #define STATE_Testing

// Internal FE key definitions:
enum {
	FE_KEY_QUIT = 1,
	FE_KEY_THRUST,
	FE_KEY_LEFT,
	FE_KEY_RIGHT,
	FE_KEY_BRAKE
	};

/******************************************************************
 * STRUCTURES:
 ******************************************************************/

/******************************************************************
 * GLOBAL VARIABLES:
 ******************************************************************/

/******************************************************************
 * LOCAL FUNCTION PROTOTYPES
 ******************************************************************/
char *GenerateSpriteCode(int width, int height, int angle);
int  *GenerateEdgeData(int width, int height, int angle);
void  FE_RedrawScreen(UN_Object *first, int nX, int nY, short *ticks);
void  AddStack(int *stack, int *stacklen, int destoff, int w, int h);
void  DoStack(int *stack, int *stacklen);

/******************************************************************
 * LOCAL VARIABLES:
 ******************************************************************/
int *xlookup, *ylookup;	// Translation table between Universe coords & screen.
         				// TBD once we get that straightened out.
char keytable[KEY_TABLE_SIZE];		// 1 if scan code pressed, 0 else.
char transtable[128];	// Tells us which key #defines go with what scan codes.
char *virtscreen;		// Virtual screen
int	 DirtyStack[256*3];	// Stack of dirty rectangles to erase.
int	 DirtyStackPos;		// Where we are in that stack.
int  RedrawAllFlag = 0;	// Currently unused as we redraw all anyways.
int  *rotatedata[256];	// Edge data for rotation code.
						// Format: Header: width/height, cos/sin
						//		   Element: src off, fract, dst off, dst width.
int *stardata;			// Star data for moving stars etc.
						// Format: Header: dx, dy
						//         Element: X, Y, speed, colour
volatile int ticks = 0;	// Current tick.
int frames = 0;			// Number of frames elapsed.
// Joystick center & usage:
int glb_joy_cx, glb_joy_cy, glb_joy_enabled;

/******************************************************************
 * AddStack
 * This should probably make sure it doesn't add the same square twice.
 ******************************************************************/
void AddStack(int *stack, int *stacklen, int destoff, int w, int h)
{
	stack[*stacklen] = (int) destoff;
	stack[*stacklen+1] = (int) w;
	stack[*stacklen+2] = (int) h;
	*stacklen += 3;
	}

/******************************************************************
 * DoStack
 ******************************************************************/
void DoStack(int *stack, int *stacklen)
{
	int i1;

	for (i1 = 0; i1 < *stacklen;i1+=3) {
		FEASM_DrawSquare(&virtscreen[stack[i1]], stack[i1+1], stack[i1+2]);
		}
	*stacklen = 0;
	}

/******************************************************************
 * FE_Malloc
 *	Currently a chain to normal malloc.  Allows possible debug code
 * to be added.
 ******************************************************************/
void *FE_Malloc(unsigned long mem)
{
	void *result;

	result = malloc(mem);
	if (!result) {
		FE_DisplayError(E_MEMORY);
		}
	return(result);
	}

/******************************************************************
 * FE_Free
 *	Required complement to malloc.
 ******************************************************************/
void FE_Free(void *ptr)
{
	free(ptr);
	}

/******************************************************************
 * FE_PointerStatus
 ******************************************************************/
void FE_PointerStatus(short* x, short* y, char* b1, char* b2)
{
	char c;
	// See need.txt for why I don't think this is correct.
//	KEYHNDLR_Enable(0);
    static char ans[80];
    printf("Which button do you want the pointer on?\n");
    printf("nEw, REP, LOAD, CONT, INST, SAVE, OPT, CRED, Quit, YES, NO\n");

	if (frames && ticks)
	{
		printf("FPS: %f\n", 18.2*5*frames / ticks);
	}

/*
	c = getch();
	printf("%c, read\n", c);
    scanf("%s",ans);
*/
//        *x = NEW_btn.ulx; *y = NEW_btn.uly;
        *x = 0; *y = 0;
#if 0
    if ((ans[0] == 'N') && (ans[1] == 'E')) {
        *x = NEW_btn.ulx; *y = NEW_btn.uly;
    } else if (ans[0] == 'R') {
        *x = REP_btn.ulx; *y = REP_btn.uly;
    } else if (ans[0] == 'L') {
        *x = LOAD_btn.ulx; *y = LOAD_btn.uly;
    } else if ((ans[0] == 'C') && (ans[0] == 'O')) {
        *x = CONT_btn.ulx; *y = CONT_btn.uly;
    } else if (ans[0] == 'I') {
        *x = INST_btn.ulx; *y = INST_btn.uly;
    } else if (ans[0] == 'S') {
        *x = SAVE_btn.ulx; *y = SAVE_btn.uly;
    } else if (ans[0] == 'O') {
        *x = OPT_btn.ulx; *y = OPT_btn.uly;
    } else if ((ans[0] == 'C') && (ans[1] == 'R')) {
        *x = CRED_btn.ulx; *y = CRED_btn.uly;
    } else if (ans[0] == 'Q') {
        *x = QUIT_btn.ulx; *y = QUIT_btn.uly;
    } else if (ans[0] == 'Y') {
        *x = YES_btn.ulx; *y = YES_btn.uly;
    } else if ((ans[0] == 'N') && (ans[1] == 'O')) {
        *x = NO_btn.ulx; *y = NO_btn.uly;
    } else {
        *x = 0; *y = 0;
    }
#endif
//	memset(&keytable[128], 0, 128);
	while (1)
	{
		if (keytable[KEY_E])
		{
	        *x = NEW_btn.ulx; *y = NEW_btn.uly;
			break;
		}
		if (keytable[KEY_Q])
		{
        	*x = QUIT_btn.ulx; *y = QUIT_btn.uly;
			FE_DeInitSystem();
			FE_DisplayError(0);
			break;
		}
		if (keytable[KEY_Y])
		{
	        *x = YES_btn.ulx; *y = YES_btn.uly;
			break;
		}
		if (keytable[75])
			break;
	}
    *x += 1; *y += 1; /* So it's inside the button box */
    *b1 = 1;
    *b2 = 1;

	ticks = 0;				// To count frames accurately!
//	KEYHNDLR_Enable(TRUE);
	}

/******************************************************************
 * FE_DisplayError
 *	Takes an error number, translates to string and writes to stdout.
 * For now crashes program.
 ******************************************************************/
void FE_DisplayError(FE_ErrorNum err)
{
	switch (err) {
		case 0:
			FEASM_SetGraph(0x3);
			printf("Invalid call to FE_DisplayError\n");
			exit(0);
			break;
		case E_MEMORY:
			FEASM_SetGraph(0x3);
			printf("Failure to allocate memory\n");
			exit(E_MEMORY);
			break;
		default:
			FEASM_SetGraph(0x3);
			printf("Unknown error %d\n", err);
			exit(err);
			break;
		}
	}

/******************************************************************
 * FE_LoadGame
 ******************************************************************/
FE_ErrorNum FE_LoadGame(short)
{
	return 0;
}

/******************************************************************
 * FE_SaveGame
 ******************************************************************/
FE_ErrorNum FE_SaveGame(short)
{
	return 0;
}

/******************************************************************
 * FE_DisplayScreen
 *  Currently, sets graph mode to text & writes text..
 ******************************************************************/
void FE_DisplayScreen(FE_ScreenNum screenNum)
{
	switch (screenNum) {
		case MAINMENU:
//			FEASM_SetGraph(0x3);
			printf("At main menu!\n");
			break;
		case PAUSEMENU:
//			FEASM_SetGraph(0x3);
			printf("At pause menu!\n");
			break;
		case CONFIRMQUITDIALOG:
//			FEASM_SetGraph(0x3);
			printf("At confirm quit menu!\n");
			break;
	}
}			

/******************************************************************
 * FE_Set
 *  Shouldn't this be get????
 ******************************************************************/
FE_Ticks FE_Set()
{
	return(ticks+1);
	}

/******************************************************************
 * FE_GetTicks
 ******************************************************************/
FE_Ticks FE_GetTicks(FE_Ticks nextUpdate)
{
    return(nextUpdate - ticks);
}

/********************************************************************
 * FE_Test
 *	Does a cheap test of rotating ability:
 ********************************************************************/
void FE_Test(char *ship, int angle)
{
	char *dest;

	dest = (char *) 0xa0000;
	dest += 100*320 + 160;
	FEASM_DrawRotated(dest, ship, rotatedata[angle]); // rotatedata[angle]);
	}

/********************************************************************
 * FE_InitScreen
 *	Sets the screen res etc.
 ********************************************************************/
void FE_InitScreen()
{
	FEASM_SetGraph(0x13);
	
	// Fill dirty stack with entire screen:
	RedrawAllFlag = 1;
	}

/********************************************************************
 * FE_InitSystem
 *	The config file normally would tell us what res etc.  For now,
 * we assume 320x200.
 ********************************************************************/
FE_ErrorNum FE_InitSystem(int *, char **) 
{
	int i1;
	JOYSTICK joy;

	// Create look up tables...
	xlookup = (int *) FE_Malloc(1024 * sizeof(int));
	ylookup = (int *) FE_Malloc(768 * sizeof(int));
	if (!(ylookup || xlookup)) {
		// Error...  We still need common error recovery!
		}
	// Fill tables:  (This is not currently used until we decide on a scale...
	for (i1 = 0; i1 < 1024; i1++) {
		xlookup[i1] = (int) ((double) i1 / 1024 * 320);
		}
	for (i1 = 0; i1 < 768; i1++) {
		ylookup[i1] = (int) ((double) i1 / 768 * 200);
		}
	// Erase keypress table:
	memset(keytable, 0, KEY_TABLE_SIZE);
	// Set translation table to 1-1 correspondence.  More on this later.
	for (i1 = 0; i1 < 128; i1++) {
		transtable[i1] = (char) i1;
		}
	// Finish setting it up (By scan codes:->)
	transtable[FE_KEY_QUIT] = 1;		// Esc.
	transtable[FE_KEY_THRUST] = 72;	// Up.
	transtable[FE_KEY_LEFT] = 75;		// left.
	transtable[FE_KEY_RIGHT] = 77;		// right.
	transtable[FE_KEY_BRAKE] = 80;		// Break.

	// Allocate virtual screen:
	virtscreen = (char *) FE_Malloc(SPRITESZ * VIRTW * VIRTH);
	memset(virtscreen, 0, SPRITESZ * VIRTW * VIRTH);

	// Generate edge table:
	for (i1 = 0; i1 < 256; i1++) {
		rotatedata[i1] = GenerateEdgeData(SPRITEW, SPRITEH, i1);
		}

	// Initialize stars:
	stardata = (int *) FE_Malloc(sizeof(int) * (NUM_STAR * 4 + 2));
	stardata[0] = 0;
	stardata[1] = 0;
	FEASM_InitStars(0x1234678, NUM_STAR, stardata);

	// Grab ints etc:
	FEASM_Init(xlookup, ylookup);
	KEYHNDLR_Install(keytable);
	FE_InitScreen();
	TIMER_Install((int *)&ticks);
	TIMER_SetRate(5);

	// Check for joystick & find center:
	JOYSTICK_GetStatus(&joy);
	glb_joy_cx = joy.ax;
	glb_joy_cy = joy.ay;

	// See if timed out:
	if (glb_joy_cx == 4000 && glb_joy_cy == 4000)
		glb_joy_enabled = 0;
	else
		glb_joy_enabled = 1;

	JOYSTICK_SetMax(max(glb_joy_cx, glb_joy_cy) + JOYBOUND*3/2);
	
	return(E_NOERROR);
	}

/********************************************************************
 * FE_DeInitSystem
 *	Frees memory, kills vectors.
 ********************************************************************/
void FE_DeInitSystem() 
{
	int i1;

	KEYHNDLR_Remove();
	FEASM_SetGraph(0x3);
	free(xlookup);
	free(ylookup);
	free(virtscreen);
	for (i1 = 0; i1 < 256; i1++) {
		free(rotatedata[i1]);
		}
	FEASM_ShutDown();
	}

/********************************************************************
 * FE_CompileSprite
 *   Takes a 128x128x256c (No colour look up for now...)
 * image & converts it to our own internal size.
 * Need a better sprite memory manager, as each sprite needs to be 256 wide,
 * ie: interleave 4 32 wide sprites together or something similar.
 ********************************************************************/
char *FE_CompileSprite(char *source)
{
	char *dest, *temp;
	int	  i1, i2;

	dest = (char *) FE_Malloc(256 * SPRITEH);
	if (!dest) {
		return(dest);
		}
	temp = dest;
	for (i1 = 0; i1 < SPRITEH; i1++) {
		for (i2 = 0; i2 < SPRITEW; i2++) {
			*temp++ = *source;
	  //		source += 4;
	  		source++;
			}
		temp += 256 - SPRITEW;
//		source += 128 - SPRITEW;	// get to next line...
		}
	return(dest);
	}

/********************************************************************
 * FE_IsKey
 *	Tells whether a given key code has been pressed.
 * Basically, looks up in table of key codes for the scan code requested,
 * and checks if that scan code has been pressed.
 ********************************************************************/
int FE_IsKey(char key)
{
	return(keytable[transtable[key]]);
	}

/******************************************************************
 * FE_GetKeys
 *  Calls FE_IsKey to build action variable
 ******************************************************************/
UN_Actions FE_GetKeys()
{
	int result = 0;

	// What are the codes???
	if (FE_IsKey(FE_KEY_QUIT))
		result |= SYS_QUIT;
	if (FE_IsKey(FE_KEY_THRUST))
		result |= ACT_THRUST;
	if (FE_IsKey(FE_KEY_LEFT))
		result |= ACT_LEFT;
	if (FE_IsKey(FE_KEY_RIGHT))
		result |= ACT_RIGHT;
	if (FE_IsKey(FE_KEY_BRAKE))
		result |= ACT_BRAKE;

	if (glb_joy_enabled)
	{
		JOYSTICK	joy;

		JOYSTICK_GetStatus(&joy);
		if (joy.ay < glb_joy_cy - JOYBOUND)
			result |= ACT_THRUST;
		if (joy.ay > glb_joy_cy + JOYBOUND)
			result |= ACT_BRAKE;
		if (joy.ax < glb_joy_cx - JOYBOUND)
			result |= ACT_LEFT;
		if (joy.ax > glb_joy_cx + JOYBOUND)
			result |= ACT_RIGHT;
		if (!joy.buttons[0])
			result |= ACT_FIRE;
		if (!joy.buttons[1])
			result |= ACT_EAT;
		if (!joy.buttons[2])
			result |= ACT_BREED;
	}

	return( result );
	}

/******************************************************************
 * FE_GetTicks
 * TBD!
 ******************************************************************/
FE_Ticks FE_GetTicks()
{
	return(-1);
	}

/******************************************************************
 * FE_ToggleSound
 ******************************************************************/
void FE_ToggleSound()
{
	return;
	}

/******************************************************************
 * FE_Update
 * TBD: Move star code into here...
 ******************************************************************/
FE_Ticks FE_Update(UN_Object* **Subsectors, long Player_X, long Player_Y,
				unsigned short sect_shiftx,
				unsigned short sect_shifty)
{
	short ticks = 0;
	unsigned int subx, suby, nextsubx, nextsuby;		// Which subsector...
	static int oX = -1, oY = -1;
	int nX, nY, dltX, dltY;

	frames++;

	// First, clear dirty stack:
	DoStack(DirtyStack, &DirtyStackPos);
	
//	printf("Player X %10d, Player Y %10d", Player_X, Player_Y);
	// Find new upper left...
	// We now redraw all stars.
	nX = Player_X - VIRTW * SPRITEW * (1 << (sect_shiftx - 10)) / 2;
	nY = Player_Y - VIRTH * SPRITEH * (1 << (sect_shifty - 10)) / 2;
	if (oX == -1) oX = nX;
	if (oY == -1) oY = nY;
	dltX = oX - nX;
	if (dltX > 65536*32)		// This constant from unobject.c
		dltX -= 65536*64;
	if (dltX < -65536*32)
		dltX += 65536*64;
	dltY = nY - oY;			// Reversed to match UN... oY - nY;
	if (dltY > 65536*32)
		dltY -= 65536*64;
	if (dltY < -65536*32)
		dltY += 65536*64;
	stardata[0] = (dltX) << 8;	// Stars move only as fast as stationary
	stardata[1] = (dltY) << 8;	// Objects.. Remove, cool effect :>
	FEASM_MoveStars(virtscreen, NUM_STAR, stardata);
	oX = nX;
	oY = nY;

	// Call redraw: Find upper left:
	subx = nX;
	suby = nY;
	subx >>= sect_shiftx;
	suby >>= sect_shifty;
	subx &= 63;					// THIS CONSTANT FROM UN_OBJECT.C!!
	suby &= 63;
	FE_RedrawScreen(Subsectors[subx][suby], Player_X, Player_Y, &ticks );
	// Is left side different?
	nextsubx = Player_X + VIRTW * SPRITEW * (1 << (sect_shiftx - 10)) / 2;
	nextsubx >>= sect_shiftx;
	nextsubx &= 63;
	if (nextsubx != subx) {
		FE_RedrawScreen(Subsectors[nextsubx][suby], Player_X, Player_Y, &ticks );
		}
	nextsuby = Player_Y + VIRTH * SPRITEH * (1 << (sect_shiftx - 10)) / 2;
	nextsuby >>= sect_shifty;
	nextsuby &= 63;
	if (nextsuby != suby) {
		FE_RedrawScreen(Subsectors[subx][nextsuby], Player_X, Player_Y, &ticks );
		}
	if ((nextsuby != suby) && (nextsubx != subx)) {
		FE_RedrawScreen(Subsectors[nextsubx][nextsuby], Player_X, 
						Player_Y, &ticks );
		}
	// Okay, backdrop drawn, ships blitted, now actually blit the screen:
	FEASM_BlitScreen((char *) 0xa0000, &virtscreen[SPRITEH*VIRTW*SPRITEW + SPRITEW], 
					VIRTW*SPRITEW-320);
	return( ticks );
	}

/********************************************************************
 * FE_RedrawScreen
 *	This has yet to be finalized.  Takes a linked list of objects to
 * read (And a rather pathetic object definition to boot)
 ********************************************************************/
void FE_RedrawScreen(UN_Object *first, int nX, int nY, short *ticks)
{
	int offx, offy;
	UN_Object *cur;
	int oX, oY;
	int dltX, dltY;

	oX = nX;
	oY = nY;

	// Now, draw all the ships
	for (cur = first; cur; cur = cur->next_sect) {
		// Get position of this object:
		offx = rotatedata[cur->facedir & 255][0]/2+1;
		offy = rotatedata[cur->facedir & 255][1]/2+1;

		dltX = cur->x - oX;
		if (dltX > 65536*32)		// This constant from unobject.c
			dltX -= 65536*64;
		if (dltX < -65536*32)
			dltX += 65536*64;
//		dltY = cur->y - oY;
		dltY = oY - cur->y;			// This is reversed to allow for 
									// correct behaviour.
		if (dltY > 65536*32)
			dltY -= 65536*64;
		if (dltY < -65536*32)
			dltY += 65536*64;
		nX = (dltX * 1024) / (64*1024) + 192;
		nY = (dltY * 1024) / (64*1024) + 132;
//printf("Object screen pos %d, %d\n", nX, nY);
		if ((nX > offx) && (nX < VIRTW*SPRITEW - offx) &&
		     (nY > offy) && (nY < VIRTH*SPRITEH - offy)) {
			AddStack(DirtyStack, &DirtyStackPos, (nX-1)+(nY-1)*VIRTW*SPRITEW, 
			         rotatedata[cur->facedir & 255][0] + 2, 
					 rotatedata[cur->facedir & 255][1] + 2);	// Remove off by one error!
			FEASM_DrawRotated(&virtscreen[nX+nY*VIRTW*SPRITEW],
							  ((AS_Picture_s *)cur->pic)->curpic, 
							  rotatedata[cur->facedir & 255]);
			}
		}
	// All drawn...  Normally retrieve tick count, but that timer not yet
	// added.
	*ticks = 0;
	}

/********************************************************************
 * GenerateEdgeData
 * 	 This will generate the appropriate edge data for the rotate code.
 ********************************************************************/
int *GenerateEdgeData(int width, int height, int angle)
{
	int dxx, dxy, dyx, dyy;		// Deltas, (dest:source)
	int sx, sy;					// Source position: 16:16.
	int osx, osy;				// Source position: 16:16, original
	int dx, dy;					// destination position: normal.
	int dw, dh;					// destination width/height.
	int i1, i2;					// Scratch.
	int *code, *ret;
	int codelen = 0;
	int edioff = 0;				// Current edi offset.
	int binit = 0;

	code = (int *) FE_Malloc(64000);	// Temporary buffer.
	if (!code)
		return(code);
// Okay, first, find destination width/height.
	i1 = width;				// Remember: we draw from centre!!!
	i2 = height;
	dxx = cos(-angle*DEG_TO_RAD)*65536;
	dxy = sin(-angle*DEG_TO_RAD)*65536;
	dyx = dxy;
	dyy = -dxx;
	dw = (abs(i1*dxx) + abs(i2*dxy)) >> 16;
	dh = (abs(i1*dyx) + abs(i2*dyy)) >> 16;		// Find widths of dest.
// Write header:
	code[codelen++] = dw;
	code[codelen++] = dh;
	code[codelen++] = dxx;
	code[codelen++] = dxy;		// Cos/Sin values used for stepping over map.
// Init sx, sy.
	osx = -dw/2 * cos(-angle*DEG_TO_RAD)*65536 - 
			dh/2 * sin(-angle*DEG_TO_RAD)*65536 + (width/2 * 65536);
	osy = -dw/2 * sin(-angle*DEG_TO_RAD)*65536 + 
			dh/2 * cos(-angle*DEG_TO_RAD)*65536 + (height/2 * 65536);
	for (dy = -dh/2; dy < -dh/2 + dh; dy++) {
		sx = osx;
		sy = osy;
		for (dx = 0; dx < dw && !POINT_IN((sx>>16), (sy>>16), width, height); dx++) {
			sx += dxx;
			sy += dxy;
			}
		// Write src offset
		code[codelen++] = (sx >> 16);
		// Write fractional source offset: Contains real sy value as well!!!
		code[codelen++] = ((sx & 0xffff) << 16) + ((sy >> 8) & 0xffff);
		// write dest offset:
		code[codelen++] = dx;
		for (; dx < dw && POINT_IN((sx>>16), (sy>>16), width, height); dx++) {
			sx += dxx;
			sy += dxy;
			}
		// write width:
		code[codelen] = dx - code[codelen-1];
		codelen++;
		osx += dyx;
		osy += dyy;
		}			// End for dy=-dh/2..dh/2

	if (codelen > 64000/sizeof(int)) {
		// This should NOT happen!!!
		}

	// Create smaller version to return:
	ret = (int *) FE_Malloc(codelen*sizeof(int));
	if (!ret) {
		free(code);
		return(NULL);
		}
	memcpy(ret, code, codelen*sizeof(int));
	free(code);
	return(ret);
	}
