/***************************************************************************\
* Alife Game - main.c code file                                             *
* ------------------------------------------------------------------------- *
* Copyright (C) 1995 - 1997 Project Von Neumann Team.  All Rights Reserved. *
* ------------------------------------------------------------------------- *
* Last Revised: 4/8/97                                                      *
* By:           Adrian Tymes                                                *
* Version #:    0.81                                                        *
* Note:         What the program does when not playing a game               *
\***************************************************************************/

#include "main.h"
#include "unobject.h"
#include "ungame.h"
#include "ai.h"
#include "fe.h"
#include "as.h"
#include "stdlib.h" /* for exit() */

/* Description:
 * #define to create button boundaries.
 *   b        = button name
 *   ulx, uly = upper left corner x and y coordinates
 *    xw,  yh = width and height of button
 * All on points 0-1023, 0-767 screen */

#define BUTTON(b, ulx,uly,xw,yh) \
	 MA_Button b##_btn = { ulx, uly, ulx+xw, uly+yh }


/* Buttons:
 * NEW   - New Game
 * LOAD  - Load Game
 * SAVE  - Save Game     NOTE: Disabled unless game paused
 * CONT  - Continue Game NOTE: Disabled unless game paused
 * REP   - Replay Movie  NOTE: "Show Background Story" when game paused
 * INST  - Instructions
 * OPT   - Options
 * CRED  - Credits
 * QUIT  - Quit
 * YES   - Yes           NOTE: Same boundaries as OPT
 * NO    - Yes           NOTE: Same boundaries as QUIT
 *
 * Layout:
 *
 *    NEW   REP   LOAD
 *
 *    CONT  INST  SAVE
 *
 *    OPT   CRED  QUIT
 *
*/
/* NOTE: If add button here also need to externally declare it in main.h */

BUTTON(NEW,  200, 150, 207, 155);

BUTTON(CONT, 200, 306, 207, 155);

BUTTON(OPT,  200, 462, 207, 155);

BUTTON(REP,  408, 150, 207, 155);

BUTTON(INST, 408, 306, 207, 155);

BUTTON(CRED, 408, 462, 207, 155);

BUTTON(LOAD, 616, 150, 207, 155);

BUTTON(SAVE, 616, 306, 207, 155);

BUTTON(QUIT, 616, 462, 207, 155);


/* Game slot layout:
 *
 *    SLOT1   SLOT2   SLOT3
 *
 *    SLOT4   SLOT5   SLOT6
 *
 *    SLOT7   SLOT8  CANCEL
 *
*/

BUTTON(SLOT1,  200, 150, 207, 155);

BUTTON(SLOT2, 200, 306, 207, 155);

BUTTON(SLOT3,  200, 462, 207, 155);

BUTTON(SLOT4,  408, 150, 207, 155);

BUTTON(SLOT5, 408, 306, 207, 155);

BUTTON(SLOT6, 408, 462, 207, 155);

BUTTON(SLOT7, 616, 150, 207, 155);

BUTTON(SLOT8, 616, 306, 207, 155);

BUTTON(CANCEL, 616, 462, 207, 155);


/* Confirm screen layout (XXXXX = no button present):
 *
 *    XXXXX   XXXXX   XXXXX
 *
 *    XXXXX   XXXXX   XXXXX
 *
 *     YES    XXXXX    NO
 *
*/

BUTTON(YES, 616, 150, 207, 155);

BUTTON(NO, 616, 462, 207, 155);


/* Description:
 * Returns TRUE only if coordinates x, y are within boundaries a,b,c,d */

#define MA_Within(x,y,a,b,c,d) ((x>a)&&(x<c)&&(y>b)&&(y<d))


/* Description:
 * Returns TRUE if the x,y is within the button 'b' boundaries */

#define MA_CheckButton(b,x,y) \
	 MA_Within(x, y, b##_btn.ulx, b##_btn.uly, b##_btn.lrx, b##_btn.lry)


/* NOTE:
 * There is a split between
 * - MA (main) functions, which will run and then dump the user back to
 *   a menu.
 * - UN (universe) functions, which deal with the game universe itself. */

/*****************************
 * Global Variables
 ****************************/

short curmovie;
					     
/*****************************
 * Function Definitions
 ****************************/

/* Description:
 * This function is called when a game is loaded from the main menu.*/

static short MA_LoadGame(short mode) {
	char b1, b2;
	short x, y;
	short slot_chosen;
	FE_DisplayScreen(LOADGAMESCREEN);
	while (1) {
		slot_chosen = 0;
		b1 = 0;
		while (!b1) {
			FE_PointerStatus(&x, &y, &b1, &b2);
		}
		/* find if a slot has been clicked on */
		if (MA_CheckButton(SLOT1, x, y)) slot_chosen = 1;
		else if (MA_CheckButton(SLOT2, x, y)) slot_chosen = 2;
		else if (MA_CheckButton(SLOT3, x, y)) slot_chosen = 3;
		else if (MA_CheckButton(SLOT4, x, y)) slot_chosen = 4;
		else if (MA_CheckButton(SLOT5, x, y)) slot_chosen = 5;
		else if (MA_CheckButton(SLOT6, x, y)) slot_chosen = 6;
		else if (MA_CheckButton(SLOT7, x, y)) slot_chosen = 7;
		else if (MA_CheckButton(SLOT8, x, y)) slot_chosen = 8;
		else if (MA_CheckButton(CANCEL, x, y)) return(mode);
		if (slot_chosen != 0) {
/* FE_DisplayScreen will have to take into account the slot
 * number called, in case that machine's implementation calls up
 * any stats about the slot being overwritten/loaded for the
 * confirmation screen */
			FE_DisplayScreen(CONFIRMLOAD + slot_chosen);
			b2 = 0;
			while (!b2) {
			/* use b2 to signal when a choice has been made */
				b1 = 0;
				while (!b1) {
					FE_PointerStatus(&x, &y, &b1, &b2);
				}
				if (MA_CheckButton(YES, x, y)) {
					if (x = FE_LoadGame(slot_chosen))
						FE_DisplayError(x);
					/* return to paused mode */
					else return(2);
				} else if (MA_CheckButton(NO, x, y)) {
					slot_chosen = 0;
					FE_DisplayScreen(LOADGAMESCREEN);
					b2 = 1;
				} else b2 = 0;
			} /* end b2 loop */
		} /* end if a slot is chosen */
	 } /* end while */
} /* end function */


/* Description:
 * This function is called when a game is saved from the main menu, which
 * is only possible when the game is paused first. */

static void MA_SaveGame() {
	char b1, b2;
	short x, y;
	short slot_chosen;
	FE_DisplayScreen(SAVEGAMESCREEN);
	while (1) {
		slot_chosen = 0;
		b1 = 0;
		while (!b1) {
			FE_PointerStatus(&x, &y, &b1, &b2);
		}
		/* find if a slot has been clicked on */
		if (MA_CheckButton(SLOT1, x, y)) slot_chosen = 1;
		else if (MA_CheckButton(SLOT2, x, y)) slot_chosen = 2;
		else if (MA_CheckButton(SLOT3, x, y)) slot_chosen = 3;
		else if (MA_CheckButton(SLOT4, x, y)) slot_chosen = 4;
		else if (MA_CheckButton(SLOT5, x, y)) slot_chosen = 5;
		else if (MA_CheckButton(SLOT6, x, y)) slot_chosen = 6;
		else if (MA_CheckButton(SLOT7, x, y)) slot_chosen = 7;
		else if (MA_CheckButton(SLOT8, x, y)) slot_chosen = 8;
		else if (MA_CheckButton(CANCEL, x, y)) return;
		if (slot_chosen != 0) {
/* FE_DisplayScreen will have to take into account the slot
 * number called, in case that machine's implementation calls up
 * any stats about the slot being overwritten/loaded for the
 * confirmation screen */
			FE_DisplayScreen(CONFIRMSAVE + slot_chosen);
			b2 = 0;
			while (!b2) {
			/* use b2 to signal when a choice has been made */
				b1 = 0;
				while (!b1) {
					FE_PointerStatus(&x, &y, &b1, &b2);
				}
				if (MA_CheckButton(YES, x, y)) {
					if (x = FE_SaveGame(slot_chosen))
						FE_DisplayError(x);
					else return;
				} else if (MA_CheckButton(NO, x, y)) {
					FE_DisplayScreen(SAVEGAMESCREEN);
					b2 = 1;
				} else b2 = 0;
			} /* end b2 loop */
		} /* end if a slot is chosen */
	 } /* end while */
} /* end function */


/* Description:
 * Calls up the in-game modifiable options menu. This will only relate
 * to those options which are similar across all platforms, since this is
 * a MA_ routine, although it will set several variables that FE will use.
 * Comments:
 * Many options will be possible, especially for universe and alife.
 * This will therefore require careful thought. */

static void MA_Options()
{
	FE_DisplayScreen(OPTIONSCREEN);
}


/* Description:
 * This function presents a dialog that prompts the user to confirm that
 * the user wants to quit the game. If the user confirms, this function
 * returns 0, otherwise it returns the input parameter.
 * Comments:
 * FE_PointerStatus is assumed to update the pointer's screen
 * representation. */

static short MA_ConfirmQuit(short mode) {
    char b1, b2;
    short x, y;
    FE_DisplayScreen(CONFIRMQUITDIALOG);
    while (1) {
	b1 = 0;
	while (!b1) {
	    FE_PointerStatus(&x, &y, &b1, &b2);
	}
	if (MA_CheckButton(YES, x, y)) return 0;
	else if (MA_CheckButton(NO, x, y)) return mode;
    }
}


/* Description:
 * - Initializes tables
 * - Initializes front end (displays error if unable to)
 * - Initializes sound (but this may be done in front end already)
 * - Shows intro movie
 * - Handles main menu
 */

int main(int argc, char *argv[]) {
	FE_ErrorNum err;
	short go;
	short x, y;
	char b1, b2;

	/* Initialize tables, etc. */
	UN_Init();

	/* Initializes front end */
	/* Certain ports may have own command line options e.g. X version
	 * FE_ will remove any options processed and update argc and argv. MA_
	 * can then check any options common to all platforms */
	if (err = FE_InitSystem(&argc, argv)) { /* contains AS_Init() call */
		  FE_DisplayError(err);
		  exit(1);
	}

	/* show intro movie */
	AS_ShowMovie(INTRO);
	go = 1;

	/* from here on, go contains the game's state:
	 * 0=quitting
	 * 1=main menu
	 * 2=paused
	 * 3=playing
	 */

	while (go) {
		if (go == 1) {
			curmovie = 0;
			FE_DisplayScreen(MAINMENU);
		} else FE_DisplayScreen(PAUSEMENU);

		b1 = 0;
		while (!b1) {
		/* This also updates the pointer's screen representation */
				FE_PointerStatus(&x, &y, &b1, &b2);
		}

		if (MA_CheckButton(NEW, x, y)) {
		/* if new game button is pressed */
			UN_InitGame();
			go = 3;
			while (go == 3) {
				go = UN_GameLoop(); /* play the game */
				if (go == 0) go = MA_ConfirmQuit(go);
/* if the user hit "quit", but selects "no" while confirming, resume the game */
			}
		} else if (MA_CheckButton(LOAD, x, y)) {
		/* if load game button is pressed */
			go = MA_LoadGame(go);
		} else if ((go == 2) && MA_CheckButton(SAVE, x, y)) {
		/* if game is paused and save game button is pressed */
			MA_SaveGame();
		} else if ((go == 2) && MA_CheckButton(CONT, x, y)) {
		/* if game is paused and continue game button is pressed */
			/* resume game */
			go = 3;
			while (go == 3) {
				go = UN_GameLoop(); /* play the game */
				if (go == 0) go = MA_ConfirmQuit(go);
/* if the user hit "quit", but selects "no" while confirming, resume the game */
			}
		} else if (MA_CheckButton(REP, x, y)) {
		/* if game replay movie button is pressed */
			if (go == 2) AS_ShowMovie(WAVE_BASE + curmovie);
			    /* Show current wave's movie if game is paused */
			else AS_ShowMovie(INTRO);
		} else if (MA_CheckButton(INST, x, y)) {
		/* if instructions button is pressed */
			AS_ShowMovie(INSTRUCTIONS);
/* No matter what format they are actually in, the universe code can think of
 * the instructions as a "movie". */
		} else if (MA_CheckButton(OPT, x, y)) {
		/* if options button is pressed */
			MA_Options();
		} else if (MA_CheckButton(CRED, x, y)) {
		/* if credits button is pressed */
			AS_ShowMovie(CREDITS);
/* As with the instructions, the credits could be anything, but the universe
 * code will still call them a "movie". */
		} else if (MA_CheckButton(QUIT, x, y)) {
		  /* if quit button is pressed */
			go = MA_ConfirmQuit(go);
		}
	 } /* end of main WHILE loop */

	 /* show exit movie */
	 AS_ShowMovie(OUTRO);
	 /* prepare game for shutdown */
	 FE_DeInitSystem();
	 /* Bye! */
	 return 0;
}
