#ifndef UNAL_H
#define UNAL_H
/************************************************************************\
* Project Von Neumann - unal.h header file                               *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: July '96                                                 *
* By          : Stuart McDonald                                          *
* Version #   : 0.76                                                     *
* Description : Header file Universe to AL system                        *
\************************************************************************/ 

#include "untypes.h"

/***************************************************************************\
* Description: Function Prototypes
\***************************************************************************/

/* Store a reference to a newly created object
 * obj->what must already be set.
*/
void UN_AddObjectRef(UN_Object* obj);

/* Remove the reference to the newly destroyed object
 * obj->what must still be valid.
*/
void UN_DeleteObjectRef(UN_Object* obj);

/* Routines called from AL to return info about object
*/
unsigned short AL_GetObjWidth(AL_ObjectRef ref);
unsigned short AL_GetObjHeight(AL_ObjectRef ref);

UN_ObjType AL_GetObjType(AL_ObjectRef ref);

UN_Velocity AL_GetAbsXVel(AL_ObjectRef ref);
UN_Velocity AL_GetAbsYVel(AL_ObjectRef ref);
UN_Velocity AL_GetAbsVelSqrd(AL_ObjectRef ref);
UN_Velocity AL_GetRelXVel(AL_ObjectRef askee, AL_ObjectRef asker);
UN_Velocity AL_GetRelYVel(AL_ObjectRef askee, AL_ObjectRef asker);
UN_Velocity AL_GetRelVelSqrd(AL_ObjectRef askee, AL_ObjectRef asker);

UN_Coord AL_GetAbsXPos(AL_ObjectRef ref);
UN_Coord AL_GetAbsYPos(AL_ObjectRef ref);
UN_Coord AL_GetRelXPos(AL_ObjectRef askee, AL_ObjectRef asker);
UN_Coord AL_GetRelYPos(AL_ObjectRef askee, AL_ObjectRef asker);

UN_Direction AL_GetAbsFacing(AL_ObjectRef ref);
UN_Direction AL_GetRelFacing(AL_ObjectRef askee, AL_ObjectRef asker);
UN_Direction AL_GetAbsHeading(AL_ObjectRef ref);
UN_Direction AL_GetRelHeading(AL_ObjectRef askee, AL_ObjectRef asker);
UN_Direction AL_GetRelDirection(AL_ObjectRef askee, AL_ObjectRef asker);

long           ALI_GetShield(AL_ObjectRef ref);
long           ALI_GetEnergy(AL_ObjectRef ref);
long           ALI_GetMass(AL_ObjectRef ref);
unsigned long  ALI_GetAge(AL_ObjectRef ref);
UN_AnimateInfo ALI_GetAnim(AL_ObjectRef ref);
UN_Direction   ALI_GetTurnRate(AL_ObjectRef ref);
unsigned long  ALI_GetThrustRate(AL_ObjectRef ref);
UN_Actions     ALI_GetActions(AL_ObjectRef ref);

#endif
